class UnexpectedCapabilities(Exception):
    pass


class ActorListenBreaker(Exception):
    pass


class ActorPoisoned(ActorListenBreaker):
    pass


class NotEnoughResources(Exception):
    pass


class NotEnoughResourcesToContinue(NotEnoughResources):
    pass



from itertools import chain, product

import pytest

from atqo import parallel_map
from atqo.distributed_apis import DIST_API_MAP


def add2(x):
    return x + 2


def extstr(x):
    return f"{x}-add"


def div(x):
    return 10 / x


@pytest.mark.parametrize(
    ["fun", "inl"],
    product([add2, extstr, div], [[1, 2, 3, 4], ["a", 10, "b", 3], [None, 2, 0]]),
)
def test_batch(fun, inl):
    res = []
    exs = []
    for x in inl:
        try:
            res.append(fun(x))
        except Exception as e:
            exs.append(e)

    map_outs = chain(
        *[
            [parallel_map(fun, inl, dapi), parallel_map(fun, iter(inl), dapi)]
            for dapi in DIST_API_MAP.keys()
        ]
    )
    for mout in map_outs:
        mres = []
        mex = []
        for o in mout:
            if isinstance(o, Exception):
                mex.append(o)
            else:
                mres.append(o)
        assert sorted(mres) == sorted(res)
        for _w in [str, type]:
            assert set(map(_w, mex)) == set(map(_w, exs))


def test_extras():

    with pytest.raises(TypeError):
        parallel_map(
            add2,
            [1, 2, 3, "X", "Y"],
            batch_size=3,
            pbar=True,
            raise_errors=True,
        )



import asyncio
import itertools
import uuid
from enum import Enum
from itertools import chain
from queue import Empty, Queue
from threading import Thread
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, List, Type

from structlog import get_logger

from atqo.bases import TaskPropertyBase

from .distributed_apis import DEFAULT_DIST_API_KEY, get_dist_api
from .exceptions import ActorListenBreaker, ActorPoisoned, NotEnoughResourcesToContinue
from .exchange import CapsetExchange
from .resource_handling import Capability, CapabilitySet, NumStore

if TYPE_CHECKING:
    from .bases import ActorBase, DistAPIBase  # pragma: no cover


POISON_KEY = frozenset([])  # just make sure it comes before any other
POISON_PILL = None
ALLOWED_CONSUMER_FAILS = 5


def _start_loop(loop: asyncio.AbstractEventLoop) -> None:
    asyncio.set_event_loop(loop)
    loop.run_forever()


class Scheduler:
    def __init__(
        self,
        actor_dict: Dict[CapabilitySet, Type["ActorBase"]],
        resource_limits: Dict[Enum, float],
        distributed_system: str = DEFAULT_DIST_API_KEY,
        reorganize_after_every_task: bool = False,  # overkill
        verbose=False,
    ) -> None:
        """Core scheduler class

        default reorganize when:
          - new tasks are added
          - no new task can be consumed
          -

        """

        self._result_queue = Queue()
        self._active_async_tasks = set()
        self._task_queues: Dict[CapabilitySet, TaskQueue] = {}

        self._loop = asyncio.new_event_loop()

        Thread(target=_start_loop, args=(self._loop,), daemon=True).start()

        self._frequent_reorg = reorganize_after_every_task
        self._verbose = verbose
        self._dist_api: DistAPIBase = get_dist_api(distributed_system)()

        self._actor_sets: Dict[CapabilitySet, ActorSet] = {}
        self._run(self._add_actor_sets(actor_dict))

        self._capset_exchange = CapsetExchange(actor_dict.keys(), resource_limits)

        # TODO
        # concurrent_task_limit: Callable[[List[TaskPropertyBase]], bool]
        # self._active_task_properties = ActiveTaskPropertySet()
        # self._task_limiter = concurrent_task_limit  # TODO

    def __del__(self):
        try:
            self._dist_api.join()
        except AttributeError:  # pragma: no cover
            pass

    def process(
        self,
        batch_producer: Callable[[], List["SchedulerTask"]],
        result_processor=None,
        min_queue_size: int = 0,
    ):

        while True:
            is_done = self.is_idle
            (result_processor or list)(self.get_processed_results())
            next_batch = batch_producer()
            batch_size = len(next_batch)
            empty_batch = batch_size == 0
            self._log("new batch", size=batch_size, was_done=is_done)
            if is_done and empty_batch:
                break
            if empty_batch:
                self.wait_until_n_tasks_remain(0)
                continue
            self.refill_task_queue(next_batch)
            try:
                self.wait_until_n_tasks_remain(min_queue_size)
            except KeyboardInterrupt:  # pragma: no cover
                self._log(f"Interrupted waiting for {self}")
                break

    def refill_task_queue(self, task_batch: Iterable["SchedulerTask"]):
        self._run(self._refill_task_queue(task_batch))

    def wait_until_n_tasks_remain(self, remaining_tasks: int = 0):
        self._run(self._await_until(remaining_tasks))

    def join(self):
        self.wait_until_n_tasks_remain(0)
        self._run(self._drain_all_actor_sets())
        try:
            self._run(asyncio.wait(self._all_actors))
        except AssertionError:
            pass
        self._run(self._cleanup())
        self._dist_api.join()

    def get_processed_results(self) -> Iterable:
        while True:
            try:
                yield self._result_queue.get(False)
            except Empty:
                break

    @property
    def is_empty(self) -> bool:
        return self.is_idle and self._result_queue.empty()

    @property
    def is_idle(self) -> bool:
        return not self._active_async_tasks

    @property
    def queued_task_count(self):
        return sum([tq.size for tq in self._task_queues.values()])

    def _run(self, coro, wait=True):
        fut = asyncio.run_coroutine_threadsafe(coro, self._loop)
        if wait:
            fut.result()

    def _log(self, logstr, **kwargs):
        if self._verbose:
            get_logger(
                api=type(self._dist_api).__name__,
                queued=self.queued_task_count,
                working=self._running_consumer_count,
            ).info(logstr, **kwargs)

    def _q_of_new_capset(self, capset: CapabilitySet) -> asyncio.Queue:

        new_task_queue = TaskQueue()
        self._task_queues[capset] = new_task_queue
        for task_cs, task_queue in self._task_queues.items():
            if task_cs > capset:
                task_queue.reset_ping()
        return new_task_queue

    async def _add_actor_sets(self, actor_dict):
        self._actor_sets = {
            capset: ActorSet(
                actor_cls,
                self._dist_api,
                capset,
                self._task_queues,
                self._verbose,
            )
            for capset, actor_cls in actor_dict.items()
        }

    async def _refill_task_queue(self, task_batch: Iterable["SchedulerTask"]):
        for scheduler_task in task_batch:
            await self._add_task(scheduler_task)
        await self._reorganize_actors()

    async def _add_task(self, scheduler_task: "SchedulerTask"):
        coro = self._await_future_and_put_result_to_queue(scheduler_task)
        async_task = self._loop.create_task(coro)
        self._active_async_tasks.add(async_task)
        capset = scheduler_task.requirements
        q = self._task_queues.get(capset) or self._q_of_new_capset(capset)
        await q.put(scheduler_task)

    async def _await_future_and_put_result_to_queue(
        self, scheduler_task: "SchedulerTask"
    ):
        scheduler_task.init_future()
        task_result = await scheduler_task.future
        self._result_queue.put(task_result)

    async def _reorganize_actors(self):
        """optimize actor set sizes

        target: minimize max n(tasks<=capset) / n(actors>=capset)
                for all task queue capsets
        limit: capset resource use * n_actors <=total resource avail
               for all actorset capsets

        heuristic:
        value of adding: decrease caused in  target / number possible remaining

        """
        need_dic = {cs: t.size for cs, t in self._task_queues.items()}
        new_needs = NumStore(need_dic)
        new_ideals = self._capset_exchange.set_values(new_needs)
        self._log(f"reorganizing on {need_dic}")
        self._log(f"reorganizing to {new_ideals}")

        for cs, new_ideal in new_ideals.items():
            await self._actor_sets[cs].set_running_actors_to(new_ideal)

        dead_end = self.queued_task_count and self._capset_exchange.idle

        if dead_end:
            await self._cleanup()
            await self._cancel_remaining_tasks()
            raise NotEnoughResourcesToContinue(
                f"{self.queued_task_count} remaining and no launchable actors"
            )

    async def _await_until(self, remaining_tasks: int = 0):
        return_when = (
            "FIRST_COMPLETED"
            if (self._frequent_reorg or remaining_tasks) > 0
            else "ALL_COMPLETED"
        )
        while len(self._active_async_tasks) > remaining_tasks:
            done, _ = await asyncio.wait(
                self._active_async_tasks, return_when=return_when
            )
            self._active_async_tasks.difference_update(done)
            if self._frequent_reorg:
                await self._reorganize_actors()

        await self._reorganize_actors()

    async def _drain_all_actor_sets(self):
        for actor_set in self._actor_sets.values():
            await actor_set.drain_to(0)

    async def _cleanup(self):
        for aset in self._actor_sets.values():
            aset.poison_queue.cancel()
        for t_queue in self._task_queues.values():
            t_queue.cancel()

    async def _cancel_remaining_tasks(self):
        for atask in self._active_async_tasks:
            atask.cancel()

    @property
    def _running_consumer_count(self):
        return sum([aset.running_actor_count for aset in self._actor_sets.values()])

    @property
    def _all_actors(self):
        return itertools.chain(
            *[aset.all_actor_tasks for aset in self._actor_sets.values()]
        )


class TaskQueue:
    def __init__(self) -> None:
        self.queue = asyncio.Queue()
        self.getting_task: asyncio.Task = asyncio.create_task(self.queue.get())
        self.ping = asyncio.Future()
        self.put = self.queue.put

    def reset_ping(self):
        self.ping.set_result(None)
        self.ping = asyncio.Future()

    def pop(self):
        out = self.getting_task.result()
        self.getting_task = asyncio.create_task(self.queue.get())
        return out

    @property
    def cancel(self):
        return self.getting_task.cancel

    @property
    def done(self):
        return self.getting_task.done

    @property
    def size(self):
        return self.queue.qsize() + int(self.getting_task.done())

    @property
    def tasks(self):
        return [self.ping, self.getting_task]


class ActorSet:
    def __init__(
        self,
        actor_cls: Type["ActorBase"],
        dist_api: "DistAPIBase",
        capset: CapabilitySet,
        task_queues: Dict[CapabilitySet, TaskQueue],
        debug: bool,
    ) -> None:
        self.actor_cls = actor_cls
        self.dist_api = dist_api
        self.capset = capset

        self.poison_queue = TaskQueue()
        self._poisoning_done_future = asyncio.Future()
        self._task_queues = task_queues
        self._actor_listening_async_task_dict: Dict[str, asyncio.Task] = {}
        self._debug = debug

        self._async_queue_get_task_dict = ...

    def __repr__(self):
        dic_str = [f"{k}={v}" for k, v in self._log_dic.items()]
        return f"{type(self).__name__}({', '.join(dic_str)}"

    async def set_running_actors_to(self, target_count):
        if target_count < self.running_actor_count:
            await self.drain_to(target_count)
        elif target_count > self.running_actor_count:
            for _ in range(self.running_actor_count, target_count):
                await self.add_new_actor()

    async def drain_to(self, target_count: int) -> int:
        n = 0
        for _ in range(target_count, self.running_actor_count):
            n += 1
            await self.poison_queue.put(POISON_PILL)
            await self._poisoning_done_future
            self._poisoning_done_future = asyncio.Future()
        return n

    async def add_new_actor(self):
        running_actor = self.dist_api.get_running_actor(actor_cls=self.actor_cls)
        listener_name = uuid.uuid1().hex
        coroutine = self._listen(
            running_actor=running_actor,
            name=listener_name,
        )
        task = asyncio.create_task(coroutine, name=listener_name)
        self._logger.info("adding consumer", listener_task=task.get_name())
        self._actor_listening_async_task_dict[listener_name] = task

    @property
    def task_count(self):
        return sum([q.size for q in self._task_queues.values()])

    @property
    def running_actor_count(self):
        return len(self._actor_listening_async_task_dict)

    @property
    def all_actor_tasks(self):
        return self._actor_listening_async_task_dict.values()

    async def _listen(self, running_actor: "ActorBase", name: str):
        self._logger.info(
            "consumer listening",
            running=type(running_actor).__name__,
        )
        fails = 0
        while True:
            next_task = await self._get_next_task()
            try:
                fails = await self._process_task(running_actor, next_task, fails)
            except ActorListenBreaker as e:
                self._logger.info(
                    "stopping consumer",
                    reason=e,
                    running=type(running_actor).__name__,
                )
                self.dist_api.kill(running_actor)
                del self._actor_listening_async_task_dict[name]
                self._poisoning_done_future.set_result(True)
                if not isinstance(e, ActorPoisoned):
                    await self.add_new_actor()
                return

    async def _get_next_task(self) -> "SchedulerTask":
        while True:
            await asyncio.wait(
                self._wait_on_tasks,
                return_when="FIRST_COMPLETED",
            )
            for t_queue in self._sorted_queues:
                if t_queue.done():
                    return t_queue.pop()

    async def _process_task(
        self,
        running_actor: "ActorBase",
        next_task: "SchedulerTask",
        fails: int,
    ):
        if next_task is POISON_PILL:
            raise ActorPoisoned("poisoned")
        try:
            out = await self.dist_api.get_future(running_actor, next_task)
            if isinstance(out, Exception):
                raise out
            next_task.set_future(out)
            return 0
        except self.dist_api.exception as e:
            self._logger.warning(
                "Remote consumption error ",
                e=e,
                te=type(e),
            )
            if self._debug:
                self._logger.exception(e)
            next_task.fail_count += 1
            if next_task.fail_count > next_task.max_fails:
                next_task.set_future(self.dist_api.parse_exception(e))
            else:
                await self._task_queues[next_task.requirements].put(next_task)
        if fails >= ALLOWED_CONSUMER_FAILS:
            raise ActorListenBreaker(f"{fails} number of fails reached")
        return fails + 1

    @property
    def _wait_on_tasks(self):
        return chain(
            *[self._task_queues[k].tasks for k in self._task_keys],
            [self.poison_queue.getting_task],
        )

    @property
    def _sorted_queues(self):
        keys = sorted(self._task_keys)
        return reversed([self.poison_queue, *map(self._task_queues.get, keys)])

    @property
    def _task_keys(self):
        return filter(self.capset.__ge__, self._task_queues.keys())

    @property
    def _logger(self):
        return get_logger(**self._log_dic)

    @property
    def _log_dic(self):
        return {
            "actor": self.actor_cls.__name__,
            "tasks": self.task_count,
            "actors_running": self.running_actor_count,
        }


class SchedulerTask:
    def __init__(
        self,
        argument: Any,
        requirements: List[Capability] = None,
        properties: List[TaskPropertyBase] = None,
        allowed_fail_count: int = 1,
    ):
        self.argument = argument
        self.requirements = CapabilitySet(requirements or [])
        self.properties = properties or []
        self.max_fails = allowed_fail_count
        self.fail_count = 0
        self.future = None

    def init_future(self):
        self.future = asyncio.Future()

    def set_future(self, value):
        self.future.set_result(value)



from functools import partial
from itertools import islice
from multiprocessing import cpu_count

from .bases import ActorBase
from .core import Scheduler, SchedulerTask
from .distributed_apis import DEFAULT_DIST_API_KEY
from .resource_handling import Capability, CapabilitySet
from .utils import partial_wrap

_RES = "CPU"
_CAP = Capability({_RES: 1})
_Task = partial_wrap(SchedulerTask, requirements=[_CAP])


class BatchProd:
    def __init__(self, iterable, batch_size, mapper=_Task) -> None:
        self._size = batch_size
        self._it = iter(iterable)
        self._mapper = mapper

    def __call__(self):
        return [*map(self._mapper, islice(self._it, self._size))]


class ActWrap(ActorBase):
    def __init__(self, fun) -> None:
        self._f = fun

    def consume(self, task_arg):
        return self._f(task_arg)


def get_simp_scheduler(n, fun, dist_sys, verbose) -> Scheduler:
    return Scheduler(
        actor_dict={CapabilitySet([_CAP]): partial_wrap(ActWrap, fun=fun)},
        resource_limits={_RES: n},
        distributed_system=dist_sys,
        verbose=verbose,
    )


def parallel_map(
    fun,
    iterable,
    dist_api=DEFAULT_DIST_API_KEY,
    batch_size=None,
    min_queue_size=None,
    workers=None,
    raise_errors=False,
    verbose=False,
    pbar=False,
):
    nw = workers or cpu_count()
    batch_size = batch_size or nw * 5
    min_queue_size = min_queue_size or batch_size // 2

    if pbar:
        from tqdm import tqdm

        iterable = tqdm(iterable)

    scheduler = get_simp_scheduler(nw, fun, dist_api, verbose)

    out = []
    outmap = partial(_raiser, li=out) if raise_errors else out.append

    scheduler.process(
        batch_producer=BatchProd(iterable, batch_size),
        result_processor=lambda li: [*map(outmap, li)],
        min_queue_size=min_queue_size,
    )
    scheduler.join()
    return out


def _raiser(e, li):
    if isinstance(e, Exception):
        raise e
    li.append(e)



from abc import abstractmethod
from pathlib import Path
from tempfile import TemporaryDirectory
from threading import Lock
from typing import Optional

SEPARATOR = "://"


class LockStoreBase:
    @abstractmethod
    def __init__(self, initstr: str) -> None:
        pass  # pragma: no cover

    def acquire(self, key) -> Lock:
        lock = self.get(key)
        lock.acquire()
        return lock

    @abstractmethod
    def get(self, key: str) -> Lock:
        pass  # pragma: no cover


class FileLockStore(LockStoreBase):
    def __init__(self, initstr: str) -> None:

        from filelock import FileLock

        self._lock_cls = FileLock

        self.root = initstr or TemporaryDirectory().name

    def get(self, key) -> Lock:
        return self._lock_cls(self._get_path(key))

    def _get_path(self, key):
        try:
            subpath = Path(key).relative_to(Path(self.root).root)
        except ValueError:
            subpath = Path(key)
        path = Path(self.root, subpath).with_suffix(".lock")
        path.parent.mkdir(exist_ok=True, parents=True)
        return path


STORE_DICT = {
    # "local": ThreadLockStore,
    # "redis": RedisLockStore,
    # "dask": DaskLockStore,
    "file": FileLockStore,
}

DEFAULT_STORE_STR = [*STORE_DICT.keys()][-1] + SEPARATOR


def get_lock_store(store_str: Optional[str] = None) -> LockStoreBase:
    f"""get a lock store based on a string

    Parameters
    ----------
    store_str : str
       {', '.join([f'{k}{SEPARATOR}initstr' for k in STORE_DICT.keys()])}

    where initstr is passed to the matching LockStore:

    {STORE_DICT}

    Returns
    -------
    LockStoreBase
        one of {[*STORE_DICT.values()]}
    """
    prefix, initstr = (store_str or DEFAULT_STORE_STR).split(SEPARATOR)
    return STORE_DICT[prefix](initstr)



from itertools import product
from random import Random

import pytest

from atqo import Capability, CapabilitySet
from atqo.exceptions import NotEnoughResources
from atqo.exchange import CapsetExchange
from atqo.resource_handling import NumStore


@pytest.mark.parametrize(
    ["seed", "res_pop", "max_caps", "max_capsets"],
    product([10, 20, 42], ["ABCDEFGHIJKL", "XYZ"], [3, 4, 5], [6, 10, 20]),
)
def test_exchange(seed, res_pop, max_caps, max_capsets):
    rng = Random(seed)
    pop_n = len(res_pop)

    resources = rng.sample(res_pop, rng.randint(1, pop_n))
    res_n = len(resources)
    res_limits = [
        (r, int(round(rng.random() * rng.randint(10, 200), -1)) + 50) for r in resources
    ]
    capcount = rng.randint(1, max_caps)
    caps = [
        Capability(
            {
                r: int(round(rng.random() * rng.randint(10, 30), -1)) + 10
                for r in rng.sample(resources, rng.randint(1, res_n))
            }
        )
        for _ in range(capcount)
    ]
    capsets = set(
        [
            CapabilitySet(rng.sample(caps, rng.randint(1, capcount)))
            for _ in range(rng.randint(1, max_capsets))
        ]
    )

    cex = CapsetExchange(capsets, resources=res_limits)
    cex.set_values(NumStore({cs: rng.randint(0, 15) for cs in capsets}))
    assert_general(cex)


def capset_cex_factory(n_act, mul):
    capsets = [CapabilitySet([Capability({"A": 1})]) for _ in range(n_act)]
    cex = CapsetExchange(
        capsets,
        resources={"A": n_act * mul},
    )
    return capsets, cex


def assert_smooth(cex, capsets, mul, n_act, _):
    out = cex.set_values({c: 10 for c in capsets})
    assert sum(out.values()) == min(mul * n_act, 10 * n_act)
    if (mul * n_act) > (10 * n_act):
        assert max(out.values()) == min(out.values())


def assert_lopside(cex, capsets, mul, n_act, param):
    qtasks = {c: n for c, n in zip(capsets, [param, *([1] * n_act)])}
    out = cex.set_values(qtasks)

    first = [*out.values()][0]
    assert sum(out.values()) == min(mul * n_act, sum(qtasks.values()))
    assert max(out.values()) == first
    if (n_act > 1) and (mul > 1):
        assert any([first > v for v in out.values()])
    assert not any([first < v for v in out.values()])


def assert_lin(cex, capsets, mul, n_act, param):
    qtasks = {c: n for c, n in zip(capsets, range(0 + param, n_act + param))}
    out = cex.set_values(qtasks)

    last = [*out.values()][-1]
    assert sum(out.values()) == min(mul * n_act, sum(qtasks.values()))
    if (mul * n_act) > sum(qtasks.values()):
        assert max(out.values()) == last
        if (n_act > 2) and (mul > 1):
            assert any([last > v for v in out.values()])
        assert not any([last < v for v in out.values()])


def assert_general(cex: CapsetExchange):
    assert sum(cex.actors_running.values()) <= (
        sum(cex.tasks_queued.base_dict.values())
    )


def no_conseq(li):
    for e1, e2 in zip(li[:-1], li[1:]):
        if e1 == e2:
            return False
    return True


fun_combs = [
    comb
    for comb in product([assert_lin, assert_lopside, assert_smooth], repeat=3)
    if no_conseq(comb)
]


@pytest.mark.parametrize(
    ["n_act", "mul", "param", "funlist"],
    product([1, 2, 10], [1, 3, 11], [3, 10], fun_combs),
)
def test_simple_cex(n_act, mul, param, funlist):
    capsets, cex = capset_cex_factory(n_act, mul)
    for fun in funlist:
        fun(cex, capsets, mul, n_act, param)
        assert_general(cex)


def test_repr():

    cex = CapsetExchange([CapabilitySet([])], {})
    assert "0" in cex.__repr__()
    with pytest.raises(NotEnoughResources):
        cex._possible_trades


def test_no_overshooting():

    limit = {"A": 20}
    caps = [Capability({"A": 1}), Capability({"A": 1})]
    cs1 = CapabilitySet(caps[:1])
    csf = CapabilitySet(caps)
    cex = CapsetExchange([cs1, csf], limit)
    out = cex.set_values(NumStore({cs1: 6, csf: 6}))
    assert out[cs1] == 6
    assert out[csf] == 6


# test cases
# one containing both caps c


def test_combining():
    limit = {"A": 2}
    caps = [Capability({"A": 1}), Capability({"A": 1}), Capability({"A": 1})]
    cs_l = CapabilitySet([caps[0]])
    cs_r = CapabilitySet([caps[-1]])
    cs1 = CapabilitySet(caps[:-1])
    cs2 = CapabilitySet(caps[1:])
    cs_sides = CapabilitySet(cs_l | cs_r)
    cex = CapsetExchange([cs1, cs2, cs_sides], limit)
    out = cex.set_values({cs_l: 1, cs_r: 1})
    assert out
    # todo
    # assert out[cex] == 1


def test_no_overcombining():
    limit = {"A": 3}
    caps = [Capability({"A": 1}), Capability({"A": 1}), Capability({"A": 1})]
    cs1 = CapabilitySet(caps[:1])
    cs2 = CapabilitySet(caps[-1:])
    csx = CapabilitySet(caps)
    cex = CapsetExchange([cs1, cs2, csx], limit)
    out = cex.set_values({cs1: 2, cs2: 2})
    assert out[csx] == 0
    assert sum(out.values()) == 3


def test_trade_finding():
    pass



from dataclasses import dataclass
from itertools import product
from pathlib import Path
from random import Random

import pytest

from atqo.distributed_apis import DIST_API_MAP
from atqo.lock_stores import SEPARATOR, STORE_DICT, LockStoreBase, get_lock_store
from atqo.simplified_functions import parallel_map


@dataclass
class IoArg:
    path: Path
    add: int

    @classmethod
    def from_args(cls, args):
        return cls(*args)


class Writer:
    def __init__(self, lock_store: LockStoreBase) -> None:
        self._locks = lock_store

    def __call__(self, arg: IoArg):
        with self._locks.get(arg.path.as_posix()):
            i = int(arg.path.read_text())
            arg.path.write_text(str(i + arg.add))


@pytest.mark.parametrize(
    ["store_key", "size", "nfiles", "dkey"],
    product(STORE_DICT.keys(), [5, 50, 500], [2, 30], DIST_API_MAP.keys()),
)
def test_para_io(tmp_path, store_key, size, nfiles, dkey):

    rng = Random(120)

    data_dir = tmp_path / "data"
    data_dir.mkdir()
    data_files = [data_dir / f"file-{i}" for i in range(nfiles)]
    for df in data_files:
        df.write_text("0")

    lockstr = f"{store_key}{SEPARATOR}{tmp_path / 'locks'}"
    lock_store = get_lock_store(lockstr)

    args = [*product(data_files, range(size))]
    rng.shuffle(args)

    parallel_map(
        Writer(lock_store),
        map(IoArg.from_args, args),
        dkey,
    )

    for fp in data_files:
        assert int(fp.read_text()) == sum(range(size))



import pytest

from atqo.core import Scheduler, SchedulerTask
from atqo.exceptions import NotEnoughResources
from atqo.resource_handling import Capability, CapabilitySet


def test_empty_scheduler():
    scheduler = Scheduler({}, {})
    assert scheduler.is_idle
    assert scheduler.is_empty
    assert scheduler.queued_task_count == 0


def test_dead_end():

    cap1 = Capability({"A": 1})
    cap2 = Capability({"A": 2})

    scheduler = Scheduler({CapabilitySet([cap1]): None}, {"A": 1})

    with pytest.raises(NotEnoughResources):
        scheduler.refill_task_queue([SchedulerTask("x", requirements=[cap1, cap2])])



from functools import partial, update_wrapper
from operator import add, sub


def dic_merge(d1, d2, _def, fun):
    return {k: fun(d1.get(k, _def), d2.get(k, _def)) for k in [*d1, *d2]}


def partial_wrap(func, *args, **kwargs):
    return update_wrapper(partial(func, *args, **kwargs), func)


sumdict = partial(dic_merge, _def=0, fun=add)
subdict = partial(dic_merge, _def=0, fun=sub)



from collections import defaultdict
from enum import Enum
from functools import cached_property, reduce
from itertools import product
from math import gcd
from typing import Dict, Iterable, List, Tuple, Union

from .exceptions import NotEnoughResources
from .resource_handling import CapabilitySet, NumStore


class CapsetExchange:
    def __init__(
        self,
        actor_capsets: Iterable[CapabilitySet],
        resources: Union[Dict[Enum, int], Iterable[Tuple[Enum, int]]],
    ) -> None:
        """trade resources for running actors with capsets

        - find possible trades
        - when prompted, filter these to possible ones
        - select best valued one
        - execute trade, refresh possible ones, repeat
        """

        self._sources = (
            [*resources.items()] if isinstance(resources, dict) else resources
        )  # can be limited what can be used together, later
        self._actor_capsets = sorted(actor_capsets)
        self.actors_running = {cs: 0 for cs in self._actor_capsets}
        self.tasks_queued = NumStore()
        self.idle_sources = [l // self._grans[eid] for eid, l in self._sources]

    def __repr__(self) -> str:
        bases = [
            ("actors used", self.actors_running),
            ("tasks queued", self.tasks_queued),
            (
                "using resources",
                self._utilized_resources,
            ),
            ("available resources", self.idle_sources),
        ]
        descr = "\n".join(f"{k}:\t{v}" for k, v in bases)
        return f"Capset Exchange: \n{descr}\n"

    def set_values(self, new_values: Union[NumStore, Dict[Enum, int]]):
        self.tasks_queued = NumStore(new_values)
        self._execute_positive_trades()
        return self.actors_running

    @property
    def idle(self):
        return not sum(self.actors_running.values())

    def _execute_positive_trades(self):
        while True:
            state = ExchangeState(self.tasks_queued, self.actors_running)
            max_value = 0
            best_trade = None
            for trade in self._possible_trades:
                if not self._is_possible(trade):
                    continue
                trade_value = state.valuation(trade)
                if trade_value > max_value:
                    best_trade = trade
                    max_value = trade_value
            if max_value <= 0:
                break
            self._execute_trade(best_trade)

    def _is_possible(self, trade: NumStore):
        for rid, num in trade:
            if isinstance(rid, int):
                rem = self.idle_sources[rid]
            else:
                rem = self.actors_running[rid]
            if (num + rem) < 0:
                return False
        return True

    def _get_source_prices(self, capset: CapabilitySet):

        by_source = self._source_combs(capset.total_resource_use)
        if not by_source:
            msg = f"can't ever start {capset}: \n{capset.total_resource_use}"
            raise NotEnoughResources(msg)

        return [NumStore({capset: 1}) - p for p in by_source]

    def _get_barter_prices(self, capset: CapabilitySet):

        bf = BarterFinder(capset, self._actor_capsets)

        barters = []
        for resources, barter_capsets in bf.barter_pairs:
            if resources.min_value > 0:
                source_combs = self._source_combs(resources.pos_part)
                barters += [barter_capsets + sc for sc in source_combs]
            else:
                barters.append(barter_capsets)
        return barters

    def _execute_trade(self, trade: NumStore):
        for rid, num in trade:
            if isinstance(rid, int):
                self.idle_sources[rid] += num
            else:
                self.actors_running[rid] += num

    def _source_combs(self, resource_use):
        combs = []
        for res_id, res_need in resource_use:
            gran_need = res_need // self._grans[res_id]
            poss_sources = [
                NumStore({sid: gran_need})
                for sid, limit in self._sources_by_res[res_id]
                if limit >= gran_need
            ]
            combs.append(poss_sources)
        if not combs:
            return []
        return [sum(poss, NumStore({})) for poss in product(*combs)]

    @property
    def _utilized_resources(self):
        return sum(
            [cs.total_resource_use * n for cs, n in self.actors_running.items()],
            start=NumStore(),
        )

    @cached_property
    def _grans(self):
        # granularities

        res_int_lists = defaultdict(list)
        for res_id, res_int in self._sources + [
            items
            for capset in self._actor_capsets
            for items in capset.total_resource_use
        ]:
            res_int_lists[res_id].append(res_int)

        return {
            res_id: reduce(gcd, res_ints) for res_id, res_ints in res_int_lists.items()
        }

    @cached_property
    def _possible_trades(self) -> Iterable[NumStore]:
        """find all (so far most/useful) possible trades

        1. find all 1 actor -> idle resource trades
          - if one is not possible, raise/warn something
        2. find some between actor trades
        """
        all_trades = []
        for capset in self._actor_capsets:
            all_trades += self._get_source_prices(capset) + self._get_barter_prices(
                capset
            )

        return all_trades + [t * -1 for t in all_trades]

    @cached_property
    def _sources_by_res(self) -> Dict[Enum, List[Tuple[int, int]]]:
        out = defaultdict(list)
        for sid, (res_id, limit) in enumerate(self._sources):
            out[res_id].append((sid, limit))
        return out


class BarterFinder:
    def __init__(self, capset: CapabilitySet, capsets: List[CapabilitySet]) -> None:
        self._capset = capset
        self._possible_barters = [cs for cs in capsets if cs != capset]
        self.barter_pairs: List[Tuple[NumStore, NumStore]] = []
        self._get_all_barter_pairs(capset.total_resource_use)

    def _get_all_barter_pairs(self, available: NumStore, in_so_far=NumStore()):
        for capset in self._possible_barters:
            rest = available - capset.total_resource_use
            if rest.min_value >= 0:
                new_far = in_so_far + NumStore({capset: 1})
                self.barter_pairs.append((rest, new_far + NumStore({self._capset: -1})))
                self._get_all_barter_pairs(rest, new_far)


class ExchangeState:
    def __init__(self, task_dic: NumStore, actor_dic: dict):
        self.rem_acts = actor_dic.copy()
        self.holes = defaultdict(lambda: 0)
        rem_tasks = 0
        for tcs, tcount in task_dic:
            task_stack = tcount
            for acs, acount in self.rem_acts.items():
                if not (acs >= tcs):
                    continue
                if acount > task_stack:
                    self.rem_acts[acs] -= task_stack
                    task_stack = 0
                    break
                task_stack -= acount
                self.rem_acts[acs] = 0
                self.holes[acs] -= task_stack
            rem_tasks += task_stack

    def valuation(self, trade: NumStore):
        value = 0
        for id_, c in trade:
            if isinstance(id_, int):
                value += c * 1e-5
                continue
            if c > 0:
                if -self.holes.get(id_, 0) < c:
                    return 0
                value += c
            if self.rem_acts[id_] < -c:
                value += c + self.rem_acts[id_]
        return value



import asyncio
import logging
import multiprocessing as mp
from asyncio import Future
from concurrent.futures import ProcessPoolExecutor
from functools import partial, partialmethod
from typing import TYPE_CHECKING, Type

from structlog import get_logger

from .bases import ActorBase, DistAPIBase

if TYPE_CHECKING:
    from .core import SchedulerTask  # pragma: no cover

logger = get_logger()


class RayAPI(DistAPIBase):
    def __init__(self):
        import ray
        from ray.exceptions import RayError, RayTaskError

        self._exc_cls = RayError
        self._task_exc = RayTaskError
        self._ray_module = ray

        ray_specs = ray.init(
            # resources=_limitset_to_ray_init(limit_set),
            log_to_driver=False,
            logging_level=logging.WARNING,
        )
        logger.info(f"ray dashboard: http://{ray_specs.get('webui_url')}")
        logger.info("launched ray with resources", **ray.cluster_resources())
        self._running = True

    @property
    def exception(self):
        return self._exc_cls

    def join(self):
        if self._running:
            self._ray_module.shutdown()
            self._running = False

    def kill(self, actor):
        self._ray_module.wait([actor.stop.remote()])
        self._ray_module.kill(actor)

    def get_running_actor(self, actor_cls: Type["ActorBase"]) -> "ActorBase":

        # ray should get the resources here...
        # not working with partial :
        if isinstance(actor_cls, partial):
            assert not actor_cls.args, "only kwargs in partials"
            root_cls = actor_cls.func
            actor_cls = type(
                root_cls.__name__,
                (ActorBase,),
                {
                    k: (
                        v if k != "__init__" else partialmethod(v, **actor_cls.keywords)
                    )
                    for k, v in root_cls.__dict__.items()
                },
            )

        return self._ray_module.remote(actor_cls).remote()

    @staticmethod
    def get_future(actor, next_task: "SchedulerTask") -> Future:
        return asyncio.wrap_future(actor.consume.remote(next_task.argument).future())

    def parse_exception(self, e):
        return e.cause if isinstance(e, self._task_exc) else e


class SyncAPI(DistAPIBase):
    pass


class MultiProcAPI(DistAPIBase):
    def __init__(self) -> None:
        self.man = mp.Manager()

    def get_running_actor(self, actor_cls: Type["ActorBase"]) -> "ActorBase":
        return MPActorWrap(actor_cls, self.man)

    @staticmethod
    def get_future(actor, next_task: "SchedulerTask") -> Future:
        return asyncio.wrap_future(actor.consume(next_task.argument))

    def join(self):
        self.man.shutdown()


class MPActorWrap(ActorBase):
    def __init__(self, inner_actor_cls: Type["ActorBase"], man: mp.Manager):

        self._in_q = man.Queue(maxsize=1)
        self._out_q = man.Queue(maxsize=1)
        self.pool = ProcessPoolExecutor(1)
        self.proc = mp.Process(
            target=_work_mp_actor,
            args=(inner_actor_cls, self._in_q, self._out_q),
        )
        self.proc.start()

    def consume(self, task_arg):
        self.pool.submit(self._in_q.put, task_arg)
        return self.pool.submit(self._out_q.get)

    def stop(self):
        self.proc.kill()
        self.proc.join()
        self.pool.shutdown()


def _work_mp_actor(actor_cls, in_q, out_q):  # pragma: no cover
    actor = actor_cls()
    while True:
        arg = in_q.get()
        try:
            res = actor.consume(arg)
        except Exception as e:
            res = e
        out_q.put(res)


DEFAULT_DIST_API_KEY = "sync"
DEFAULT_MULTI_API = "mp"
DIST_API_MAP = {
    DEFAULT_DIST_API_KEY: SyncAPI,
    "ray": RayAPI,
    DEFAULT_MULTI_API: MultiProcAPI,
}


def get_dist_api(key) -> "DistAPIBase":
    try:
        return DIST_API_MAP[key]
    except KeyError:
        default = DIST_API_MAP[DEFAULT_DIST_API_KEY]
        err = f"unknown distributed system: {key}, defaulting {default}"
        logger.warning(err)
        return default



"""Async Task Queue Orchestrator with Complex Resource Management"""

from .bases import ActorBase, DistAPIBase, TaskPropertyBase  # noqa: F401
from .core import Scheduler, SchedulerTask  # noqa: F401
from .exceptions import UnexpectedCapabilities  # noqa: F401
from .resource_handling import Capability, CapabilitySet  # noqa: F401
from .simplified_functions import parallel_map  # noqa: F401

__version__ = "0.1.1"



from atqo import ActorBase, Capability, CapabilitySet, Scheduler, SchedulerTask
from atqo.distributed_apis import DEFAULT_DIST_API_KEY
from atqo.simplified_functions import BatchProd

LIMIT_DIC = {"A": 3}

cap1 = Capability({"A": 1})
cap2 = Capability({"A": 1})


class Actor(ActorBase):
    def consume(self, task_arg):
        return f"done-{task_arg}"


class ListProd:
    def __init__(self, list_base, batch_size) -> None:
        self._size = batch_size
        self._list = list_base
        self._i = 0

    def __call__(self):
        out = []
        for _ in range(self._size):
            try:
                out.append(self._list[self._i])
                self._i += 1
            except IndexError:
                break
        return out


def test_over_actors():

    dist_sys = DEFAULT_DIST_API_KEY
    actor_dict = {
        CapabilitySet([cap2, cap1]): Actor,
    }

    scheduler = Scheduler(
        actor_dict=actor_dict,
        resource_limits=LIMIT_DIC,
        distributed_system=dist_sys,
        verbose=True,
    )

    tasks = [
        SchedulerTask("task1", requirements=[cap1]),
    ]

    out = []

    def _proc(ol):
        for e in ol:
            out.append(e)

    scheduler.process(BatchProd(tasks, 2, lambda x: x), result_processor=_proc)
    scheduler.join()

    assert out == ["done-task1"]


def test_recurse():
    dist_sys = DEFAULT_DIST_API_KEY
    actor_dict = {
        CapabilitySet([cap2, cap1]): Actor,
    }

    scheduler = Scheduler(
        actor_dict=actor_dict,
        resource_limits=LIMIT_DIC,
        distributed_system=dist_sys,
        verbose=True,
    )

    tasks = [SchedulerTask(f"task{i}", requirements=[cap1]) for i in range(1, 9)]

    out = []

    def _proc(ol):
        size = 0
        for e in ol:
            size += 1
            out.append(e)
        for i in range(size // 2):
            tasks.append(tasks[i])

    scheduler.process(ListProd(tasks, 10), result_processor=_proc)
    scheduler.join()

    muls = [(1, 4), (2, 3), (3, 2), (4, 2), *[(i2, 1) for i2 in range(5, 9)]]

    assert sorted(out) == sum(
        [[f"done-task{i}"] * k for i, k in muls],
        [],
    )



from enum import Enum

import pytest

from atqo import ActorBase, Capability, CapabilitySet, Scheduler, SchedulerTask
from atqo.distributed_apis import DIST_API_MAP


class REnum(Enum):
    CPU = 0
    CONN = 1
    MEM = 2


LIMIT_DIC = {REnum.CPU: 6, REnum.CONN: 2000, REnum.MEM: 3500}

file_uploader = Capability({REnum.CPU: 1, REnum.CONN: 400, REnum.MEM: 1000}, name="ul")
bigfile_handling = Capability({REnum.CONN: 100, REnum.MEM: 1000}, name="big")
file_downloader = Capability({REnum.CPU: 1, REnum.CONN: 500, REnum.MEM: 750}, name="dl")


class _TestBase(ActorBase):

    prefix = "nofing"

    def __init__(self) -> None:
        self._log("TEST intited")

    def consume(self, task_arg):
        self._log(f"TEST consuming {task_arg}")
        if task_arg == "fing":
            raise ValueError("oh-oh")
        return f"{self.prefix} {task_arg}"

    def stop(self):
        self._log("TEST stopping")


class Uploader(_TestBase):
    prefix = "uploaded"


class Downloader(_TestBase):
    prefix = "downloaded"


actor_dict = {
    CapabilitySet([file_uploader, bigfile_handling]): Uploader,
    CapabilitySet([file_uploader]): Uploader,
    CapabilitySet([file_downloader, bigfile_handling]): Downloader,
    CapabilitySet([file_downloader]): Downloader,
}


class _Producer:
    def __init__(self, *task_sets) -> None:
        self.task_sets = task_sets
        self._i = -1

    def __call__(self):
        self._i += 1
        try:
            return self.task_sets[self._i]
        except IndexError:
            return []


@pytest.mark.parametrize("dist_sys", [*DIST_API_MAP.keys(), "wrong"])
def test_minor_integration(dist_sys):
    reorg = True

    scheduler = Scheduler(
        actor_dict=actor_dict,
        resource_limits=LIMIT_DIC,
        distributed_system=dist_sys,
        reorganize_after_every_task=reorg,
        verbose=True,
    )

    tasks = [
        SchedulerTask("small file", requirements=[file_uploader]),
        SchedulerTask("bigger file", requirements=[file_uploader, bigfile_handling]),
        SchedulerTask("complex", requirements=[file_downloader, bigfile_handling]),
    ]

    out = []
    errs = []

    def _processor(results):
        for r in results:
            if isinstance(r, ValueError):
                errs.append(r)
            else:
                out.append(r)

    scheduler.process(
        _Producer(tasks, tasks, [SchedulerTask("fing", [file_downloader])], tasks),
        _processor,
    )
    scheduler.join()

    assert sorted(out) == sorted(
        ["uploaded small file", "uploaded bigger file", "downloaded complex"] * 3
    )
    assert errs
    assert "Upload" in scheduler._actor_sets[CapabilitySet([file_uploader])].__repr__()



from atqo.resource_handling import Capability, NumStore


def test_num_store():

    d1 = {"A": 10}
    ns1 = NumStore(d1)
    ns2 = ns1 + NumStore({"A": 5})
    assert ns2 == NumStore({"A": 15})
    assert ns1 == (ns2 - NumStore({"A": 5}))
    assert ns1.__repr__() == d1.__repr__()
    assert ns2 > ns1 and ns2 >= ns1 and ns1 < ns2

    assert ns1.min_value == 10
    {ns1: "works in dict"}
    set([ns1, ns2])


def test_caps():
    c1 = Capability({"A": 1}, "nx")
    assert "nx" in c1.__repr__()
    assert c1 == c1
    assert c1 != Capability({"A": 1})



from abc import ABC, abstractmethod
from asyncio import Future
from functools import cached_property
from typing import TYPE_CHECKING, Type

from structlog import get_logger

if TYPE_CHECKING:
    from .core import SchedulerTask  # pragma: no cover


class TaskPropertyBase:  # pragma: no cover
    # TODO
    def __repr__(self):
        param_str = ", ".join(f"{k}={v}" for k, v in self.__dict__.items())
        return f"{type(self).__name__}({param_str})"

    def __hash__(self):
        return id(self).__hash__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()


class ActorBase(ABC):
    @abstractmethod
    def consume(self, task_arg):
        pass  # pragma: no cover

    def stop(self):
        """if any cleanup needed"""
        pass  # pragma: no cover

    @cached_property
    def _log(self):
        return get_logger(actor=type(self).__name__).info


class DistAPIBase(ABC):
    @property
    def exception(self):
        return Exception

    def join(self):
        """wait on all running tasks"""
        pass

    @staticmethod
    def kill(actor):
        actor.stop()

    @staticmethod
    def get_running_actor(actor_cls: Type["ActorBase"]) -> ActorBase:
        return actor_cls()

    @staticmethod
    def get_future(actor, next_task: "SchedulerTask") -> Future:
        f = Future()
        f.set_result(actor.consume(next_task.argument))
        return f

    @staticmethod
    def parse_exception(e):
        return e



from enum import Enum
from typing import Any, Dict, Optional, Union
from uuid import uuid4

from .utils import subdict, sumdict


class Capability:
    def __init__(
        self,
        resource_needs: Union[Dict[Enum, int], "NumStore"],
        name: Optional[str] = None,
    ):
        """capability that actors can have and tasks might need

        Parameters
        ----------
        resource_needs : Dict[Enum, int]
            need to be integers for simplicity
        name : Optional[str], optional
            identifier, by default None
        """
        self.id_ = name or uuid4().hex
        if isinstance(resource_needs, dict):
            resource_needs = NumStore(resource_needs)
        self.resource_needs = resource_needs

    def __hash__(self):
        return self.id_.__hash__()

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.id_[:3]}, {self.resource_needs})"

    def __add__(self, other: Union["Capability", "NumStore"]):
        if isinstance(other, Capability):
            other = other.resource_needs
        return Capability(self.resource_needs + other)


class ActiveTaskPropertySet:
    pass


class NumStore:
    def __init__(self, use: Optional[Dict[Any, float]] = None):
        if isinstance(use, type(self)):
            use = use.base_dict
        self.base_dict = use or {}
        self.get = self.base_dict.get

    def __eq__(self, other):
        assert isinstance(other, type(self))
        return self.base_dict == other.base_dict

    def __le__(self, other):
        assert isinstance(other, type(self))
        return all(
            [
                v <= other.base_dict.get(k, -float("inf"))
                for k, v in self.base_dict.items()
            ]
        )

    def __ge__(self, other):
        return other <= self

    def __lt__(self, other):
        return not (self == other) & (self <= other)

    def __gt__(self, other):
        return not (self == other) & (self >= other)

    def __mul__(self, other: int):
        return type(self)({k: v * other for k, v in self.base_dict.items()})

    def __add__(self, other):
        assert isinstance(other, type(self))
        return type(self)(sumdict(self.base_dict, other.base_dict))

    def __sub__(self, other):
        assert isinstance(other, type(self))
        return type(self)(subdict(self.base_dict, other.base_dict))

    def __repr__(self):
        return str(self.base_dict)

    def __hash__(self) -> int:
        return frozenset(self.base_dict.items()).__hash__()

    def __iter__(self):
        for it in self.base_dict.items():
            yield it

    @property
    def min_value(self):
        return min(self.base_dict.values())

    @property
    def pos_part(self):
        return type(self)({k: v for k, v in self if v > 0})


class CapabilitySet(frozenset):
    def __repr__(self) -> str:
        return f"Set({'-'.join([c.id_[:3] for c in self])})"

    @property
    def total_resource_use(self) -> NumStore:
        return sum(self, start=Capability({})).resource_needs
